@@ -89,51 +89,6 @@ static int class_to_call_type(int cls)
 	}
 }
 
-static unsigned int alloc_next_id(struct voicecall_data *d)
-{
-	unsigned int i;
-
-	for (i = 1; i < sizeof(d->id_list) * 8; i++) {
-		if (d->id_list & (0x1 << i))
-			continue;
-
-		d->id_list |= (0x1 << i);
-		return i;
-	}
-
-	return 0;
-}
-
-static void release_id(struct voicecall_data *d, unsigned int id)
-{
-	d->id_list &= ~(0x1 << id);
-}
-
-static gint call_compare_by_status(gconstpointer a, gconstpointer b)
-{
-	const struct ofono_call *call = a;
-	int status = GPOINTER_TO_INT(b);
-
-	if (status != call->status)
-		return 1;
-
-	return 0;
-}
-
-static gint call_compare(gconstpointer a, gconstpointer b)
-{
-	const struct ofono_call *ca = a;
-	const struct ofono_call *cb = b;
-
-	if (ca->id < cb->id)
-		return -1;
-
-	if (ca->id > cb->id)
-		return 1;
-
-	return 0;
-}
-
 static struct ofono_call *create_call(struct voicecall_data *d, int type,
 					int direction, int status,
 					const char *num, int num_type, int clip)
@@ -146,7 +101,7 @@ static struct ofono_call *create_call(struct voicecall_data *d, int type,
 	if (!call)
 		return NULL;
 
-	call->id = alloc_next_id(d);
+	call->id = at_util_alloc_next_id(&d->id_list);
 	call->type = type;
 	call->direction = direction;
 	call->status = status;
@@ -159,66 +114,11 @@ static struct ofono_call *create_call(struct voicecall_data *d, int type,
 
 	call->clip_validity = clip;
 
-	d->calls = g_slist_insert_sorted(d->calls, call, call_compare);
+	d->calls = g_slist_insert_sorted(d->calls, call, at_util_call_compare);
 
 	return call;
 }
 
-static GSList *parse_clcc(GAtResult *result)
-{
-	GAtResultIter iter;
-	GSList *l = NULL;
-	int id, dir, status, type;
-	struct ofono_call *call;
-
-	g_at_result_iter_init(&iter, result);
-
-	while (g_at_result_iter_next(&iter, "+CLCC:")) {
-		const char *str = "";
-		int number_type = 129;
-
-		if (!g_at_result_iter_next_number(&iter, &id))
-			continue;
-
-		if (!g_at_result_iter_next_number(&iter, &dir))
-			continue;
-
-		if (!g_at_result_iter_next_number(&iter, &status))
-			continue;
-
-		if (!g_at_result_iter_next_number(&iter, &type))
-			continue;
-
-		if (!g_at_result_iter_skip_next(&iter))
-			continue;
-
-		if (g_at_result_iter_next_string(&iter, &str))
-			g_at_result_iter_next_number(&iter, &number_type);
-
-		call = g_try_new0(struct ofono_call, 1);
-
-		if (!call)
-			break;
-
-		call->id = id;
-		call->direction = dir;
-		call->status = status;
-		call->type = type;
-		strncpy(call->phone_number.number, str,
-				OFONO_MAX_PHONE_NUMBER_LENGTH);
-		call->phone_number.type = number_type;
-
-		if (strlen(call->phone_number.number) > 0)
-			call->clip_validity = 0;
-		else
-			call->clip_validity = 2;
-
-		l = g_slist_insert_sorted(l, call, call_compare);
-	}
-
-	return l;
-}
-
 static void clcc_poll_cb(gboolean ok, GAtResult *result, gpointer user_data)
 {
 	struct ofono_voicecall *vc = user_data;
@@ -236,7 +136,7 @@ static void clcc_poll_cb(gboolean ok, GAtResult *result, gpointer user_data)
 		return;
 	}
 
-	calls = parse_clcc(result);
+	calls = at_util_parse_clcc(result);
 
 	n = calls;
 	o = vd->calls;
@@ -260,7 +160,7 @@ static void clcc_poll_cb(gboolean ok, GAtResult *result, gpointer user_data)
 				ofono_voicecall_disconnected(vc, oc->id,
 								reason, NULL);
 
-			release_id(vd, oc->id);
+			at_util_release_id(&vd->id_list, oc->id);
 
 			o = o->next;
 		} else if (nc && (!oc || (nc->id < oc->id))) {
@@ -373,6 +273,7 @@ static void atd_cb(gboolean ok, GAtResult *result, gpointer user_data)
 	int validity = 2;
 	struct ofono_error error;
 	struct ofono_call *call;
+	GSList *l;
 
 	dump_response("atd_cb", ok, result);
 
@@ -381,6 +282,17 @@ static void atd_cb(gboolean ok, GAtResult *result, gpointer user_data)
 	if (!ok)
 		goto out;
 
+	/* On a success, make sure to put all active calls on hold */
+	for (l = vd->calls; l; l = l->next) {
+		call = l->data;
+
+		if (call->status != 0)
+			continue;
+
+		call->status = 1;
+		ofono_voicecall_notify(vc, call);
+	}
+
 	g_at_result_iter_init(&iter, result);
 
 	if (g_at_result_iter_next(&iter, "+COLP:")) {
@@ -527,7 +439,7 @@ static void clcc_cb(gboolean ok, GAtResult *result, gpointer user_data)
 		goto out;
 	}
 
-	calls = parse_clcc(result);
+	calls = at_util_parse_clcc(result);
 
 	if (calls == NULL) {
 		CALLBACK_WITH_FAILURE(cb, 0, NULL, cbd->data);
@@ -727,9 +639,14 @@ static void ring_notify(GAtResult *result, gpointer user_data)
 
 	dump_response("ring_notify", TRUE, result);
 
+	/* See comment in CRING */
+	if (g_slist_find_custom(vd->calls, GINT_TO_POINTER(5),
+				at_util_call_compare_by_status))
+		return;
+
 	/* RING can repeat, ignore if we already have an incoming call */
 	if (g_slist_find_custom(vd->calls, GINT_TO_POINTER(4),
-				call_compare_by_status))
+				at_util_call_compare_by_status))
 		return;
 
 	/* Generate an incoming call of unknown type */
@@ -755,9 +672,20 @@ static void cring_notify(GAtResult *result, gpointer user_data)
 
 	dump_response("cring_notify", TRUE, result);
 
+	/* Handle the following situation:
+	 * Active Call + Waiting Call.  Active Call is Released.  The Waiting
+	 * call becomes Incoming and RING/CRING indications are signaled.
+	 * Sometimes these arrive before we managed to poll CLCC to find about
+	 * the stage change.  If this happens, simply ignore the RING/CRING
+	 * when a waiting call exists (cannot have waiting + incoming in GSM)
+	 */
+	if (g_slist_find_custom(vd->calls, GINT_TO_POINTER(5),
+				at_util_call_compare_by_status))
+		return;
+
 	/* CRING can repeat, ignore if we already have an incoming call */
 	if (g_slist_find_custom(vd->calls, GINT_TO_POINTER(4),
-				call_compare_by_status))
+				at_util_call_compare_by_status))
 		return;
 
 	g_at_result_iter_init(&iter, result);
@@ -802,7 +730,7 @@ static void clip_notify(GAtResult *result, gpointer user_data)
 	dump_response("clip_notify", TRUE, result);
 
 	l = g_slist_find_custom(vd->calls, GINT_TO_POINTER(4),
-				call_compare_by_status);
+				at_util_call_compare_by_status);
 
 	if (l == NULL) {
 		ofono_error("CLIP for unknown call");
@@ -866,6 +794,11 @@ static void ccwa_notify(GAtResult *result, gpointer user_data)
 
 	dump_response("ccwa_notify", TRUE, result);
 
+	/* Some modems resend CCWA, ignore it the second time around */
+	if (g_slist_find_custom(vd->calls, GINT_TO_POINTER(5),
+				at_util_call_compare_by_status))
+		return;
+
 	g_at_result_iter_init(&iter, result);
 
 	if (!g_at_result_iter_next(&iter, "+CCWA:"))
@@ -990,6 +923,9 @@ static void at_voicecall_remove(struct ofono_voicecall *vc)
 
 	g_slist_foreach(vd->calls, (GFunc) g_free, NULL);
 	g_slist_free(vd->calls);
+
+	ofono_voicecall_set_data(vc, NULL);
+
 	g_free(vd);
 }
 
